{****************************************}
{* Main unit for LogicGates application *}
{****************************************}

unit logic;

// Implementation of the interactive graphics:
//   - the circuit layout is constituted by a mixture of shapes and images
//   - the display of the logic gate is done by loading its picture from file
//   - the LEDs are colored ellipses; different color styles are used for on/off status
//   - the movement of the switches is done by changing the vertical position (TOP property) of the corresponding shapes
//   - the switch shapes being very small and difficult to click by the user, I used a transperant image, placed over them;
//     the OnClick event of the image is used to trigger the actions to be done when the corresponding switch has been clicked.

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ExtCtrls, Menus, StdCtrls, Grids;

type
  TBits = 0..1;
  TArray4 = array[0..3] of TBits;
  {*********}
  { TfLogic }
  {*********}
  TfLogic = class(TForm)
    mLogic: TMainMenu;
    mGate, mGateAND, mGateOR, mGateXOR, mGateNAND, mGateNOR, mGateXNOR, mGateExit: TMenuItem;
    mHelp, mHelpHelp, mHelpAbout: TMenuItem;
    stTitle: TStaticText;
    meGate: TMemo;
    Label1, Label2, Label3, Label4: TLabel;
    imGate, imSwitch1, imSwitch2: TImage;
    shSwitch1H, shSwitch1V, shSwitch2H, shSwitch2V: TShape;
    shLed1, shLed2, shLedOut: TShape;
    shA,  shB,  Shape6,  Shape7,  Shape8,  Shape9,  Shape10, Shape11, Shape12: TShape;
    Shape15, Shape16, Shape17, Shape18, Shape21, Shape22, Shape23, Shape24, Shape25: TShape;
    Shape26, Shape27, Shape28, Shape29, Shape30, Shape31, Shape32, Shape33, Shape34: TShape;
    StaticText1: TStaticText;
    sgTruthTable: TStringGrid;
    procedure FormCreate(Sender: TObject);
    procedure mGateANDClick(Sender: TObject);
    procedure mGateORClick(Sender: TObject);
    procedure mGateXORClick(Sender: TObject);
    procedure mGateNANDClick(Sender: TObject);
    procedure mGateNORClick(Sender: TObject);
    procedure mGateXNORClick(Sender: TObject);
    procedure mGateExitClick(Sender: TObject);
    procedure mHelpHelpClick(Sender: TObject);
    procedure mHelpAboutClick(Sender: TObject);
    procedure imSwitch1Click(Sender: TObject);
    procedure imSwitch2Click(Sender: TObject);
  private
    sOperator: string;
    aTruthTable: TArray4;
    bSwitch1, bSwitch2: TBits;
  end;

const
  LogicOperators: array[1..6] of string = (
    'and', 'or', 'xor', 'nand', 'nor', 'xnor'
  );
  GateDescriptions: array[1..6] of string = (
    'The output of an AND gate is high only when all of the inputs are high.',
    'The output of an OR gate is high when any input is high.',
    'The output of an XOR gate is high when either input is high, but not when both inputs are high.',
    'The output of an NAND gate is low only when all of the inputs are high.',
    'The output of an NOR gate is low when any input is high.',
    'The output of an XNOR gate is low when either input is high, but not when both inputs are high.'
  );
  TruthTables: array[1..6, 0..3] of TBits =
    ((0, 0, 0, 1), (0, 1, 1, 1), (0, 1, 1, 0), (1, 1, 1, 0), (1, 0, 0, 0), (1, 0, 0, 1)
  );

var
  fLogic: TfLogic;

implementation

{$R *.lfm}

{ Logic operation for given inputs and truth table }

procedure LogicOperation(A, B: TBits; var TruthTable: TArray4);

var
  TX: Integer;
  Output: TBits;

begin
  TX := A * 2 + B;                                                             // truth table index for given inputs A and B
  Output := TruthTable[TX];                                                    // logic operation result (as found in truth table)
  // Turn red LED on or off
  if Output = 0 then
    fLogic.shLedOut.Brush.Style := bsBDiagonal
  else
    fLogic.shLedOut.Brush.Style := bsSolid;
end;

{ Choose one of the 6 logic gates }

procedure LogicGate(GX: Integer; var LogicOperator: string; var TruthTable: TArray4; var Sw1, Sw2: TBits);

var
  I: Integer;
  GateFile: string;

begin
  LogicOperator := LogicOperators[GX];                                         // logic operator/gate index
  fLogic.stTitle.Caption := 'Logic gate: ' + UpperCase(LogicOperator);         // form title
  fLogic.meGate.Clear;                                                         // logic gate description
  fLogic.meGate.Lines.AddText(GateDescriptions[GX]);
  fLogic.imGate.Hint := UpperCase(LogicOperator) + ' gate';
  for I := 0 to 3 do begin                                                     // logic operator truth table
    TruthTable[I] := TruthTables[GX, I];
    fLogic.sgTruthTable.Cells[2, I + 1] := '    ' + IntToStr(TruthTable[I]);
  end;
  // Display logic gate picture
  GateFile := LogicOperator + '.jpg';
  fLogic.imGate.Picture.LoadFromFile(GateFile);
  // Picture positioning corrections
  fLogic.imGate.Left := 355; fLogic.imGate.Top := 272;
  if (LogicOperator = 'xor') or (LogicOperator = 'xnor') then begin
    fLogic.imGate.Top := 271;
    if LogicOperator = 'xnor' then
      fLogic.imGate.Left := 353;
  end;
  // Gate inputs positioning corrections
  fLogic.shA.Top := 291; fLogic.shB.Top := 320;
  fLogic.shA.Width := 61; fLogic.shB.Width := 61;
  if LogicOperator = 'and' then
    fLogic.shA.Top := 292
  else if LogicOperator = 'xor' then
    fLogic.shB.Top := 321
  else if (LogicOperator = 'nand') or (LogicOperator = 'nor') then begin
    fLogic.shA.Width := 59; fLogic.shB.Width := 59;
    if LogicOperator = 'nor' then begin
      fLogic.shA.Top := 292; fLogic.shB.Top := 321;
    end;
  end
  else if LogicOperator = 'xnor' then begin
    fLogic.shA.Top := 292;
    fLogic.shA.Width := 55; fLogic.shB.Width := 55;
  end;
  // Switches: Both open
  Sw1 := 0; Sw2 := 0;
  fLogic.shSwitch1V.Top := 216; fLogic.shSwitch1H.Top := 234;
  fLogic.shSwitch2V.Top := 328; fLogic.shSwitch2H.Top := 346;
  // Green LEDs: Both off
  fLogic.shLed1.Brush.Style := bsBDiagonal;
  fLogic.shLed2.Brush.Style := bsBDiagonal;
  // Compute logic operation for this gate with A=0 and B=0
  LogicOperation(0, 0, TruthTable);
end;

{*********}
{ TfLogic }
{*********}

{ Application start: Choose the AND gate }

procedure TfLogic.FormCreate(Sender: TObject);

begin
  fLogic.mGateAND.Click;
end;

{ Menu item "Gate > AND": Choose the AND gate }

procedure TfLogic.mGateANDClick(Sender: TObject);

begin
  LogicGate(1, sOperator, aTruthTable, bSwitch1, bSwitch2);
end;

{ Menu item "Gate > OR": Choose the OR gate }

procedure TfLogic.mGateORClick(Sender: TObject);

begin
  LogicGate(2, sOperator, aTruthTable, bSwitch1, bSwitch2);
end;

{ Menu item "Gate > XOR": Choose the XOR gate }

procedure TfLogic.mGateXORClick(Sender: TObject);

begin
  LogicGate(3, sOperator, aTruthTable, bSwitch1, bSwitch2);
end;

{ Menu item "Gate > NAND": Choose the NAND gate }

procedure TfLogic.mGateNANDClick(Sender: TObject);

begin
  LogicGate(4, sOperator, aTruthTable, bSwitch1, bSwitch2);
end;

{ Menu item "Gate > NOR": Choose the NOR gate }

procedure TfLogic.mGateNORClick(Sender: TObject);

begin
  LogicGate(5, sOperator, aTruthTable, bSwitch1, bSwitch2);
end;

{ Menu item "Gate > XNOR": Choose the XNOR gate }

procedure TfLogic.mGateXNORClick(Sender: TObject);

begin
  LogicGate(6, sOperator, aTruthTable, bSwitch1, bSwitch2);
end;

{ Menu item "Gate > Exit": Exit the application }

procedure TfLogic.mGateExitClick(Sender: TObject);

begin
  Close;
end;

{ Menu item "Help > Help": Display application help }

procedure TfLogic.mHelpHelpClick(Sender: TObject);

var
  S: string;

begin
  S := 'Choose one of the logic gates in the "Gate" menu. Click switch 1 and/or 2 to open/close ';
  S += 'it. A closed switch lets the current flow; the green LED indicates that the gate''s input ';
  S += 'is high. The red output LED goes on or off as indicated in the logical operator''s truth table.';
  MessageDlg('"LogicGates" Help"', S, mtInformation, [mbOK], 0);
end;

{ Menu item "Help > About": Display application about }

procedure TfLogic.mHelpAboutClick(Sender: TObject);

var
  S: string;

begin
  S := 'Basic logic gates circuits.' + LineEnding + LineEnding;
  S += 'Version 1.1, © allu, March 2018 - August 2021.';
  MessageDlg('About "LogicGates"', S, mtInformation, [mbOK], 0);
end;

{ Switch 1 click: Perform logic operation for switch 1 now on/off }

procedure TfLogic.imSwitch1Click(Sender: TObject);

begin
  // Switch 1 was open
  if bSwitch1 = 0 then begin
    bSwitch1 := 1;                                                             // close switch 1
    shSwitch1V.Top := 227; shSwitch1H.Top := 245;
    shLed1.Brush.Style := bsSolid;                                             // turn LED1 on
  end
  // Switch 2 was closed
  else begin
    bSwitch1 := 0;                                                             // open switch 1
    shSwitch1V.Top := 216; shSwitch1H.Top := 234;
    shLed1.Brush.Style := bsBDiagonal;                                         // turn LED1 off
  end;
  LogicOperation(bSwitch1, bSwitch2, aTruthTable);                             // perform logic operation
end;

{ Switch 2 click: Perform logic operation for switch 2 now on/off }

procedure TfLogic.imSwitch2Click(Sender: TObject);

begin
    // Switch 2 was open
  if bSwitch2 = 0 then begin
    bSwitch2 := 1;                                                             // close switch 2
    shSwitch2V.Top := 339; shSwitch2H.Top := 357;
    shLed2.Brush.Style := bsSolid;                                             // turn LED2 on
  end
    // Switch 2 was closed
  else begin
    bSwitch2 := 0;                                                             // open switch 2
    shSwitch2V.Top := 328; shSwitch2H.Top := 346;
    shLed2.Brush.Style := bsBDiagonal;                                         // turn LED2 off
  end;
  LogicOperation(bSwitch1, bSwitch2, aTruthTable);                             // perform logic operation
end;

end.

